#!/usr/bin/python
# -*- coding: utf-8 -*-

# Authors:
#   Thomas Woerner <twoerner@redhat.com>
#
# Copyright (C) 2020 Red Hat
# see file 'COPYING' for use and warranty information
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import sys
from galaxyfy import galaxyfy_playbook


def module_EXAMPLES(module_in, project_prefix, collection_prefix):
    out_lines = []

    with open(module_in) as in_f:
        changed = False
        example = False
        example_lines = []
        for line in in_f:
            stripped = line.strip()
            if stripped in ['EXAMPLES = """', "EXAMPLES = '''"]:
                example = True
                out_lines.append(line)
                continue
            elif example and stripped in ["'''", '"""']:
                _out_lines, _changed = \
                    galaxyfy_playbook(project_prefix, collection_prefix,
                                      example_lines)
                for _line in _out_lines:
                    out_lines.append(_line)
                example_lines = []
                if _changed:
                    changed = True
                example = False
                out_lines.append(line)
                continue
            if example:
                example_lines.append(line)
            else:
                out_lines.append(line)

    if changed:
        with open(module_in, "w") as out_f:
            for line in out_lines:
                out_f.write(line)


module_EXAMPLES(sys.argv[1], sys.argv[2], sys.argv[3])



#!/usr/bin/python
# -*- coding: utf-8 -*-
"""ansible-freeipa iparole module implementation."""

# Authors:
#   Rafael Guterres Jeffman <rjeffman@redhat.com>
#
# Copyright (C) 2020 Red Hat
# see file 'COPYING' for use and warranty information
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

ANSIBLE_METADATA = {
    "metadata_version": "1.0",
    "supported_by": "community",
    "status": ["preview"],
}


DOCUMENTATION = """
---
module: iparole
short description: Manage FreeIPA role
description: Manage FreeIPA role
options:
  ipaadmin_principal:
    description: The admin principal.
    default: admin
  ipaadmin_password:
    description: The admin password.
    required: false
  role:
    description: The list of role name strings.
    required: true
    aliases: ["cn"]
  description:
    description: A description for the role.
    required: false
  rename:
    description: Rename the role object.
    required: false
  user:
    description: List of users.
    required: false
  group:
    description: List of groups.
    required: false
  host:
    description: List of hosts.
    required: false
  hostgroup:
    description: List of hostgroups.
    required: false
  service:
    description: List of services.
    required: false
  action:
    description: Work on role or member level.
    choices: ["role", "member"]
    default: role
    required: false
  state:
    description: The state to ensure.
    choices: ["present", "absent"]
    default: present
    required: true
"""

EXAMPLES = """
- name: Ensure a role named `somerole` is present.
  iparole:
    ipaadmin_password: SomeADMINpassword
    name: somerole

- name: Ensure user `pinky` is a memmer of role `somerole`.
  iparole:
    ipaadmin_password: SomeADMINpassword
    name: somerole
    user:
    - pinky
    action: member

- name: Ensure a role named `somerole` is absent.
  iparole:
    ipaadmin_password: SomeADMINpassword
    name: somerole
    state: absent
"""

# pylint: disable=wrong-import-position
# pylint: disable=import-error
# pylint: disable=no-name-in-module
from ansible.module_utils.basic import AnsibleModule
from ansible.module_utils._text import to_text
from ansible.module_utils.ansible_freeipa_module import \
    temp_kinit, temp_kdestroy, valid_creds, api_connect, api_command, \
    gen_add_del_lists, compare_args_ipa, module_params_get, api_get_realm
import six


if six.PY3:
    unicode = str


def find_role(module, name):
    """Find if a role with the given name already exist."""
    try:
        _result = api_command(module, "role_show", name, {"all": True})
    except Exception:  # pylint: disable=broad-except
        # An exception is raised if role name is not found.
        return None
    else:
        return _result["result"]


def gen_args(module):
    """Generate arguments for executing commands."""
    arg_map = {
        "description": "description",
        "rename": "rename",
    }
    args = {}

    for param, arg in arg_map.items():
        value = module_params_get(module, param)
        if value is not None:
            args[arg] = value

    return args


def check_parameters(module):
    """Check if parameters passed for module processing are valid."""
    action = module_params_get(module, "action")
    state = module_params_get(module, "state")

    invalid = []

    if state == "present":
        if action == "member":
            invalid.extend(['description', 'rename'])

    if state == "absent":
        invalid.extend(['description', 'rename'])
        if action != "member":
            invalid.extend(['privilege'])

    for arg in invalid:
        if module_params_get(module, arg) is not None:
            module.fail_json(
                msg="Argument '%s' can not be used with action '%s'" %
                (arg, state))


def verify_credentials(module):
    """Ensure there are valid Kerberos credentials."""
    ccache_dir = None
    ccache_name = None

    ipaadmin_principal = module_params_get(module, "ipaadmin_principal")
    ipaadmin_password = module_params_get(module, "ipaadmin_password")

    if not valid_creds(module, ipaadmin_principal):
        ccache_dir, ccache_name = temp_kinit(ipaadmin_principal,
                                             ipaadmin_password)

    return (ccache_dir, ccache_name)


def member_intersect(module, attr, memberof, res_find):
    """Filter member arguments from role found by intersection."""
    params = module_params_get(module, attr)
    if not res_find:
        return params
    filtered = []
    if params:
        existing = res_find.get(memberof, [])
        filtered = list(set(params) & set(existing))
    return filtered


def member_difference(module, attr, memberof, res_find):
    """Filter member arguments from role found by difference."""
    params = module_params_get(module, attr)
    if not res_find:
        return params
    filtered = []
    if params:
        existing = res_find.get(memberof, [])
        filtered = list(set(params) - set(existing))
    return filtered


def ensure_absent_state(module, name, action, res_find):
    """Define commands to ensure absent state."""
    commands = []

    if action == "role":
        commands.append([name, 'role_del', {}])

    if action == "member":

        members = member_intersect(
            module, 'privilege', 'memberof_privilege', res_find)
        if members:
            commands.append([name, "role_remove_privilege",
                             {"privilege": members}])

        member_args = {}
        for key in ['user', 'group', 'host', 'hostgroup']:
            items = member_intersect(
                        module, key, 'member_%s' % key, res_find)
            if items:
                member_args[key] = items

        _services = filter_service(module, res_find,
                                   lambda res, svc: res.startswith(svc))
        if _services:
            member_args['service'] = _services

        # Only add remove command if there's at least one member no manage.
        if member_args:
            commands.append([name, "role_remove_member", member_args])

    return commands


def filter_service(module, res_find, predicate):
    """
    Filter service based on predicate.

    Compare service name with existing ones matching
    at least until `@` from principal name.

    Predicate is a callable that accepts the existing service, and the
    modified service to be compared to.
    """
    _services = []
    service = module_params_get(module, 'service')
    if service:
        existing = [to_text(x) for x in res_find.get('member_service', [])]
        for svc in service:
            svc = svc if '@' in svc else ('%s@' % svc)
            found = [x for x in existing if predicate(x, svc)]
            _services.extend(found)
    return _services


def ensure_role_with_members_is_present(module, name, res_find):
    """Define commands to ensure member are present for action `role`."""
    commands = []
    privilege_add, privilege_del = gen_add_del_lists(
        module_params_get(module, "privilege"),
        res_find.get('memberof_privilege', []))

    if privilege_add:
        commands.append([name, "role_add_privilege",
                         {"privilege": privilege_add}])
    if privilege_del:
        commands.append([name, "role_remove_privilege",
                         {"privilege": privilege_del}])

    add_members = {}
    del_members = {}

    for key in ["user", "group", "host", "hostgroup"]:
        add_list, del_list = gen_add_del_lists(
            module_params_get(module, key),
            res_find.get('member_%s' % key, [])
        )
        if add_list:
            add_members[key] = add_list
        if del_list:
            del_members[key] = [to_text(item) for item in del_list]

    service = [
        to_text(svc) if '@' in svc else ('%s@%s' % (svc, api_get_realm()))
        for svc in (module_params_get(module, 'service') or [])
    ]
    existing = [str(svc) for svc in res_find.get('member_service', [])]
    add_list, del_list = gen_add_del_lists(service, existing)
    if add_list:
        add_members['service'] = add_list
    if del_list:
        del_members['service'] = [to_text(item) for item in del_list]

    if add_members:
        commands.append([name, "role_add_member", add_members])
    if del_members:
        commands.append([name, "role_remove_member", del_members])

    return commands


def ensure_members_are_present(module, name, res_find):
    """Define commands to ensure members are present for action `member`."""
    commands = []

    members = member_difference(
        module, 'privilege', 'memberof_privilege', res_find)
    if members:
        commands.append([name, "role_add_privilege",
                         {"privilege": members}])

    member_args = {}
    for key in ['user', 'group', 'host', 'hostgroup']:
        items = member_difference(
                    module, key, 'member_%s' % key, res_find)
        if items:
            member_args[key] = items

    _services = filter_service(module, res_find,
                               lambda res, svc: not res.startswith(svc))
    if _services:
        member_args['service'] = _services

    if member_args:
        commands.append([name, "role_add_member", member_args])

    return commands


def process_command_failures(command, result):
    """Process the result of a command, looking for errors."""
    # Get all errors
    # All "already a member" and "not a member" failures in the
    # result are ignored. All others are reported.
    errors = []
    if "failed" in result and len(result["failed"]) > 0:
        for item in result["failed"]:
            failed_item = result["failed"][item]
            for member_type in failed_item:
                for member, failure in failed_item[member_type]:
                    if "already a member" in failure \
                       or "not a member" in failure:
                        continue
                    errors.append("%s: %s %s: %s" % (
                        command, member_type, member, failure))
    return errors


def process_commands(module, commands):
    """Process the list of IPA API commands."""
    errors = []
    exit_args = {}
    changed = False
    for name, command, args in commands:
        try:
            result = api_command(module, command, name, args)
            if "completed" in result:
                if result["completed"] > 0:
                    changed = True
            else:
                changed = True

            errors = process_command_failures(command, result)
        except Exception as exception:  # pylint: disable=broad-except
            module.fail_json(
                msg="%s: %s: %s" % (command, name, str(exception)))

    if errors:
        module.fail_json(msg=", ".join(errors))

    return changed, exit_args


def role_commands_for_name(module, state, action, name):
    """Define commands for the Role module."""
    commands = []

    rename = module_params_get(module, "rename")

    res_find = find_role(module, name)

    if state == "present":
        args = gen_args(module)

        if action == "role":
            if res_find is None:
                if rename is not None:
                    module.fail_json(msg="Cannot `rename` inexistent role.")
                commands.append([name, 'role_add', args])
                res_find = {}
            else:
                if not compare_args_ipa(module, args, res_find):
                    commands.append([name, 'role_mod', args])

        if action == "member":
            if res_find is None:
                module.fail_json(msg="No role '%s'" % name)

        cmds = ensure_role_with_members_is_present(module, name, res_find)
        commands.extend(cmds)

    if state == "absent" and res_find is not None:
        cmds = ensure_absent_state(module, name, action, res_find)
        commands.extend(cmds)

    return commands


def create_module():
    """Create module description."""
    ansible_module = AnsibleModule(
        argument_spec=dict(
            # generalgroups
            ipaadmin_principal=dict(type="str", default="admin"),
            ipaadmin_password=dict(type="str", required=False, no_log=True),

            name=dict(type="list", aliases=["cn"], default=None,
                      required=True),
            # present
            description=dict(required=False, type="str", default=None),
            rename=dict(required=False, type="str", default=None),

            # members
            privilege=dict(required=False, type='list', default=None),
            user=dict(required=False, type='list', default=None),
            group=dict(required=False, type='list', default=None),
            host=dict(required=False, type='list', default=None),
            hostgroup=dict(required=False, type='list', default=None),
            service=dict(required=False, type='list', default=None),

            # state
            action=dict(type="str", default="role",
                        choices=["role", "member"]),
            state=dict(type="str", default="present",
                       choices=["present", "absent"]),
        ),
        supports_check_mode=True,
        mutually_exclusive=[],
        required_one_of=[]
    )

    ansible_module._ansible_debug = True  # pylint: disable=protected-access

    return ansible_module


def main():
    """Process role module script."""
    ansible_module = create_module()
    check_parameters(ansible_module)

    # Init
    ccache_dir = None
    ccache_name = None
    try:
        ccache_dir, ccache_name = verify_credentials(ansible_module)
        api_connect()

        state = module_params_get(ansible_module, "state")
        action = module_params_get(ansible_module, "action")
        names = module_params_get(ansible_module, "name")
        commands = []

        for name in names:
            cmds = role_commands_for_name(ansible_module, state, action, name)
            commands.extend(cmds)

        changed, exit_args = process_commands(ansible_module, commands)

    except Exception as exception:  # pylint: disable=broad-except
        ansible_module.fail_json(msg=str(exception))

    finally:
        temp_kdestroy(ccache_dir, ccache_name)

    # Done
    ansible_module.exit_json(changed=changed, **exit_args)


if __name__ == "__main__":
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Authors:
#   Thomas Woerner <twoerner@redhat.com>
#
# Based on ipa-replica-install code
#
# Copyright (C) 2018  Red Hat
# see file 'COPYING' for use and warranty information
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import print_function

ANSIBLE_METADATA = {
    'metadata_version': '1.0',
    'supported_by': 'community',
    'status': ['preview'],
}

DOCUMENTATION = '''
---
module: ipareplica_setup_http
short description: Setup HTTP
description:
  Setup HTTP
options:
  setup_ca:
    description: Configure a dogtag CA
    required: yes
  setup_kra:
    description: Configure a dogtag KRA
    required: yes
  no_pkinit:
    description: Disable pkinit setup steps
    required: yes
  no_ui_redirect:
    description: Do not automatically redirect to the Web UI
    required: yes
  subject_base:
    description:
      The certificate subject base (default O=<realm-name>).
      RDNs are in LDAP order (most specific RDN first).
    required: no
  config_master_host_name:
    description: The config master_host_name setting
    required: no
  config_ca_host_name:
    description: The config ca_host_name setting
    required: no
  ccache:
    description: The local ccache
    required: no
  _ca_enabled:
    description: The installer _ca_enabled setting
    required: yes
  _ca_file:
    description: The installer _ca_file setting
    required: yes
  _http_pkcs12_info:
    description: The installer _http_pkcs12_info setting
    required: yes
  _top_dir:
    description: The installer _top_dir setting
    required: no
  dirman_password:
    description: Directory Manager (master) password
    required: no
author:
    - Thomas Woerner
'''

EXAMPLES = '''
'''

RETURN = '''
'''

import os
import inspect

from ansible.module_utils.basic import AnsibleModule
from ansible.module_utils.ansible_ipa_replica import (
    AnsibleModuleLog, setup_logging, installer, DN, paths, sysrestore,
    gen_env_boostrap_finalize_core, constants, api_bootstrap_finalize,
    gen_ReplicaConfig, gen_remote_api, api, redirect_stdout, create_ipa_conf,
    install_http
)


def main():
    ansible_module = AnsibleModule(
        argument_spec=dict(
            # server
            setup_ca=dict(required=False, type='bool'),
            setup_kra=dict(required=False, type='bool'),
            no_pkinit=dict(required=False, type='bool'),
            no_ui_redirect=dict(required=False, type='bool'),
            # certificate system
            subject_base=dict(required=True),
            config_master_host_name=dict(required=True),
            config_ca_host_name=dict(required=True),
            ccache=dict(required=True),
            _ca_enabled=dict(required=False, type='bool'),
            _ca_file=dict(required=False),
            _http_pkcs12_info=dict(required=False, type='list'),
            _top_dir=dict(required=True),
            dirman_password=dict(required=True, no_log=True),
        ),
        supports_check_mode=True,
    )

    ansible_module._ansible_debug = True
    setup_logging()
    ansible_log = AnsibleModuleLog(ansible_module)

    # get parameters #

    options = installer
    options.setup_ca = ansible_module.params.get('setup_ca')
    options.setup_kra = ansible_module.params.get('setup_kra')
    options.no_pkinit = ansible_module.params.get('no_pkinit')
    options.no_ui_redirect = ansible_module.params.get('no_ui_redirect')
    # certificate system
    options.subject_base = ansible_module.params.get('subject_base')
    if options.subject_base is not None:
        options.subject_base = DN(options.subject_base)
    # additional
    master_host_name = ansible_module.params.get('config_master_host_name')
    ca_host_name = ansible_module.params.get('config_master_host_name')
    ccache = ansible_module.params.get('ccache')
    os.environ['KRB5CCNAME'] = ccache
    # os.environ['KRB5CCNAME'] = ansible_module.params.get('installer_ccache')
    # installer._ccache = ansible_module.params.get('installer_ccache')
    ca_enabled = ansible_module.params.get('_ca_enabled')
    http_pkcs12_info = ansible_module.params.get('_http_pkcs12_info')
    options._top_dir = ansible_module.params.get('_top_dir')
    dirman_password = ansible_module.params.get('dirman_password')

    # init #

    fstore = sysrestore.FileStore(paths.SYSRESTORE)

    ansible_log.debug("== INSTALL ==")

    promote = installer.promote

    env = gen_env_boostrap_finalize_core(paths.ETC_IPA,
                                         constants.DEFAULT_CONFIG)
    api_bootstrap_finalize(env)
    config = gen_ReplicaConfig()
    config.subject_base = options.subject_base
    config.dirman_password = dirman_password
    config.setup_ca = options.setup_ca
    config.master_host_name = master_host_name
    config.ca_host_name = ca_host_name
    config.promote = installer.promote

    remote_api = gen_remote_api(master_host_name, paths.ETC_IPA)
    # installer._remote_api = remote_api

    conn = remote_api.Backend.ldap2
    ccache = os.environ['KRB5CCNAME']

    # There is a api.Backend.ldap2.connect call somewhere in ca, ds, dns or
    # ntpinstance
    api.Backend.ldap2.connect()
    conn.connect(ccache=ccache)

    cafile = paths.IPA_CA_CRT
    with redirect_stdout(ansible_log):
        ansible_log.debug("-- INSTALL_HTTP --")

        # We need to point to the master when certmonger asks for
        # HTTP certificate.
        # During http installation, the HTTP/hostname principal is created
        # locally then the installer waits for the entry to appear on the
        # master selected for the installation.
        # In a later step, the installer requests a SSL certificate through
        # Certmonger (and the op adds the principal if it does not exist yet).
        # If xmlrpc_uri points to the soon-to-be replica,
        # the httpd service is not ready yet to handle certmonger requests
        # and certmonger tries to find another master. The master can be
        # different from the one selected for the installation, and it is
        # possible that the principal has not been replicated yet. This
        # may lead to a replication conflict.
        # This is why we need to force the use of the same master by
        # setting xmlrpc_uri
        create_ipa_conf(fstore, config, ca_enabled,
                        master=config.master_host_name)

        argspec = inspect.getargspec(install_http)
        if "promote" in argspec.args:
            install_http(
                config,
                auto_redirect=not options.no_ui_redirect,
                promote=promote,
                pkcs12_info=http_pkcs12_info,
                ca_is_configured=ca_enabled,
                ca_file=cafile)
        else:
            if "fstore" not in argspec.args:
                install_http(
                    config,
                    auto_redirect=not options.no_ui_redirect,
                    pkcs12_info=http_pkcs12_info,
                    ca_is_configured=ca_enabled,
                    ca_file=cafile)
            else:
                install_http(
                    config,
                    auto_redirect=not options.no_ui_redirect,
                    pkcs12_info=http_pkcs12_info,
                    ca_is_configured=ca_enabled,
                    ca_file=cafile,
                    fstore=fstore)

        # Need to point back to ourself after the cert for HTTP is obtained
        create_ipa_conf(fstore, config, ca_enabled)

    # done #

    ansible_module.exit_json(changed=True)


if __name__ == '__main__':
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Authors:
#   Thomas Woerner <twoerner@redhat.com>
#
# Based on ipa-client-install code
#
# Copyright (C) 2017  Red Hat
# see file 'COPYING' for use and warranty information
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

__all__ = ["gssapi", "version", "ipadiscovery", "api", "errors", "x509",
           "constants", "sysrestore", "certmonger", "certstore",
           "delete_persistent_client_session_data", "ScriptError",
           "CheckedIPAddress", "validate_domain_name", "normalize_hostname",
           "validate_hostname", "services", "tasks", "CalledProcessError",
           "write_tmp_file", "ipa_generate_password", "DN", "kinit_keytab",
           "kinit_password", "GSSError", "CLIENT_INSTALL_ERROR",
           "is_ipa_client_installed", "CLIENT_ALREADY_CONFIGURED",
           "nssldap_exists", "remove_file", "check_ip_addresses",
           "print_port_conf_info", "configure_ipa_conf", "purge_host_keytab",
           "configure_sssd_conf", "realm_to_suffix", "run", "timeconf",
           "serialization", "configure_krb5_conf", "get_ca_certs",
           "SECURE_PATH", "get_server_connection_interface",
           "disable_ra", "client_dns",
           "configure_certmonger", "update_ssh_keys",
           "configure_openldap_conf", "hardcode_ldap_server",
           "get_certs_from_ldap", "save_state", "create_ipa_nssdb",
           "configure_nisdomain", "configure_ldap_conf",
           "configure_nslcd_conf", "nosssd_files", "configure_ssh_config",
           "configure_sshd_config", "configure_automount",
           "configure_firefox", "sync_time", "check_ldap_conf",
           "sssd_enable_ifp"]

from ipapython.version import NUM_VERSION, VERSION

if NUM_VERSION < 30201:
    # See ipapython/version.py
    IPA_MAJOR, IPA_MINOR, IPA_RELEASE = [int(x) for x in VERSION.split(".", 2)]
    IPA_PYTHON_VERSION = IPA_MAJOR*10000 + IPA_MINOR*100 + IPA_RELEASE
else:
    IPA_PYTHON_VERSION = NUM_VERSION


class installer_obj(object):
    def __init__(self):
        pass

    def set_logger(self, logger):
        self.logger = logger

    # def __getattribute__(self, attr):
    #    value = super(installer_obj, self).__getattribute__(attr)
    #    if not attr.startswith("--") and not attr.endswith("--"):
    #        logger.debug(
    #            "  <-- Accessing installer.%s (%s)" % (attr, repr(value)))
    #    return value

    # def __getattr__(self, attr):
    #    # logger.info("  --> ADDING missing installer.%s" % attr)
    #    self.logger.warn("  --> ADDING missing installer.%s" % attr)
    #    setattr(self, attr, None)
    #    return getattr(self, attr)

    # def __setattr__(self, attr, value):
    #    logger.debug("  --> Setting installer.%s to %s" % (attr, repr(value)))
    #    return super(installer_obj, self).__setattr__(attr, value)

    def knobs(self):
        for name in self.__dict__:
            yield self, name


# Initialize installer settings
installer = installer_obj()
# Create options
options = installer
options.interactive = False
options.unattended = not options.interactive

if NUM_VERSION >= 40400:
    # IPA version >= 4.4

    import sys
    import inspect
    import gssapi
    import logging

    from ipapython import version
    try:
        from ipaclient.install import ipadiscovery
    except ImportError:
        from ipaclient import ipadiscovery
    from ipalib import api, errors, x509
    from ipalib import constants
    try:
        from ipalib import sysrestore
    except ImportError:
        try:
            from ipalib.install import sysrestore
        except ImportError:
            from ipapython import sysrestore
    try:
        from ipalib.install import certmonger
    except ImportError:
        from ipapython import certmonger
    try:
        from ipalib.install import certstore
    except ImportError:
        from ipalib import certstore
    from ipalib.rpc import delete_persistent_client_session_data
    from ipapython import certdb, ipautil
    from ipapython.admintool import ScriptError
    from ipapython.ipautil import CheckedIPAddress
    from ipalib.util import validate_domain_name, normalize_hostname, \
        validate_hostname
    from ipaplatform import services
    from ipaplatform.paths import paths
    from ipaplatform.tasks import tasks
    try:
        from cryptography.hazmat.primitives import serialization
    except ImportError:
        serialization = None
    from ipapython.ipautil import CalledProcessError, write_tmp_file, \
        ipa_generate_password
    from ipapython.dn import DN
    try:
        from ipalib.install.kinit import kinit_keytab, kinit_password
    except ImportError:
        from ipapython.ipautil import kinit_keytab, kinit_password
    from ipapython.ipa_log_manager import standard_logging_setup
    from gssapi.exceptions import GSSError
    try:
        from ipaclient.install.client import configure_krb5_conf, \
            get_ca_certs, SECURE_PATH, get_server_connection_interface, \
            disable_ra, client_dns, \
            configure_certmonger, update_ssh_keys, configure_openldap_conf, \
            hardcode_ldap_server, get_certs_from_ldap, save_state, \
            create_ipa_nssdb, configure_ssh_config, configure_sshd_config, \
            configure_automount, configure_firefox, configure_nisdomain, \
            CLIENT_INSTALL_ERROR, is_ipa_client_installed, \
            CLIENT_ALREADY_CONFIGURED, nssldap_exists, remove_file, \
            check_ip_addresses, print_port_conf_info, configure_ipa_conf, \
            purge_host_keytab, configure_sssd_conf, configure_ldap_conf, \
            configure_nslcd_conf, nosssd_files
        get_ca_cert = None
    except ImportError:
        # Create temporary copy of ipa-client-install script (as
        # ipa_client_install.py) to be able to import the script easily
        # and also to remove the global finally clause in which the
        # generated ccache file gets removed. The ccache file will be
        # needed in the next step.
        # This is done in a temporary directory that gets removed right
        # after ipa_client_install has been imported.
        import shutil
        import tempfile
        temp_dir = tempfile.mkdtemp(dir="/tmp")
        sys.path.append(temp_dir)
        temp_file = "%s/ipa_client_install.py" % temp_dir

        with open("/usr/sbin/ipa-client-install", "r") as f_in:
            with open(temp_file, "w") as f_out:
                for line in f_in:
                    if line.startswith("finally:"):
                        break
                    f_out.write(line)
        import ipa_client_install

        shutil.rmtree(temp_dir, ignore_errors=True)
        sys.path.remove(temp_dir)

        argspec = inspect.getargspec(ipa_client_install.configure_krb5_conf)
        if argspec.keywords is None:
            def configure_krb5_conf(
                    cli_realm, cli_domain, cli_server, cli_kdc, dnsok,
                    filename, client_domain, client_hostname, force=False,
                    configure_sssd=True):
                global options
                options.force = force
                options.sssd = configure_sssd
                return ipa_client_install.configure_krb5_conf(
                    cli_realm, cli_domain, cli_server, cli_kdc, dnsok, options,
                    filename, client_domain, client_hostname)
        else:
            configure_krb5_conf = ipa_client_install.configure_krb5_conf
        if NUM_VERSION < 40100:
            get_ca_cert = ipa_client_install.get_ca_cert
            get_ca_certs = None
        else:
            get_ca_cert = None
            get_ca_certs = ipa_client_install.get_ca_certs
        SECURE_PATH = ("/bin:/sbin:/usr/kerberos/bin:/usr/kerberos/sbin:"
                       "/usr/bin:/usr/sbin")

        get_server_connection_interface = \
            ipa_client_install.get_server_connection_interface
        disable_ra = ipa_client_install.disable_ra
        client_dns = ipa_client_install.client_dns
        configure_certmonger = ipa_client_install.configure_certmonger
        update_ssh_keys = ipa_client_install.update_ssh_keys
        configure_openldap_conf = ipa_client_install.configure_openldap_conf
        hardcode_ldap_server = ipa_client_install.hardcode_ldap_server
        get_certs_from_ldap = ipa_client_install.get_certs_from_ldap
        save_state = ipa_client_install.save_state

        create_ipa_nssdb = certdb.create_ipa_nssdb

        argspec = inspect.getargspec(ipa_client_install.configure_nisdomain)
        if len(argspec.args) == 3:
            configure_nisdomain = ipa_client_install.configure_nisdomain
        else:
            def configure_nisdomain(options, domain, statestore=None):
                return ipa_client_install.configure_nisdomain(options, domain)

        configure_ldap_conf = ipa_client_install.configure_ldap_conf
        configure_nslcd_conf = ipa_client_install.configure_nslcd_conf
        nosssd_files = ipa_client_install.nosssd_files

        configure_ssh_config = ipa_client_install.configure_ssh_config
        configure_sshd_config = ipa_client_install.configure_sshd_config
        configure_automount = ipa_client_install.configure_automount
        configure_firefox = ipa_client_install.configure_firefox

    from ipapython.ipautil import realm_to_suffix, run

    try:
        from ipaclient.install import timeconf
        time_service = "chronyd"
    except ImportError:
        try:
            from ipaclient.install import ntpconf as timeconf
        except ImportError:
            from ipaclient import ntpconf as timeconf
        time_service = "ntpd"

    try:
        from ipaclient.install.client import sync_time
    except ImportError:
        sync_time = None

    try:
        from ipaclient.install.client import check_ldap_conf
    except ImportError:
        check_ldap_conf = None

    try:
        from ipaclient.install.client import sssd_enable_ifp
    except ImportError:
        sssd_enable_ifp = None

    logger = logging.getLogger("ipa-client-install")
    root_logger = logger

else:
    # IPA version < 4.4

    raise Exception("freeipa version '%s' is too old" % VERSION)


def setup_logging():
    standard_logging_setup(
        paths.IPACLIENT_INSTALL_LOG, verbose=False, debug=False,
        filemode='a', console_format='%(message)s')


def ansible_module_get_parsed_ip_addresses(ansible_module,
                                           param='ip_addresses'):
    ip_addresses = ansible_module.params.get(param)
    if ip_addresses is None:
        return None

    ip_addrs = []
    for ip in ip_addresses:
        try:
            ip_parsed = ipautil.CheckedIPAddress(ip)
        except Exception as e:
            ansible_module.fail_json(msg="Invalid IP Address %s: %s" % (ip, e))
        ip_addrs.append(ip_parsed)
    return ip_addrs



# Authors:
#   Florence Blanc-Renaud <frenaud@redhat.com>
#
# Copyright (C) 2017  Red Hat
# see file 'COPYING' for use and warranty information
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

try:
    import gssapi
except ImportError:
    gssapi = None
import os
import shutil
import subprocess
import tempfile
from jinja2 import Template

from ansible.errors import AnsibleError
from ansible.module_utils._text import to_native
from ansible.plugins.action import ActionBase


def run_cmd(args, stdin=None):
    """Execute an external command."""
    p_in = None
    p_out = subprocess.PIPE
    p_err = subprocess.PIPE

    if stdin:
        p_in = subprocess.PIPE

    p = subprocess.Popen(args, stdin=p_in, stdout=p_out, stderr=p_err,
                         close_fds=True)
    __temp, stderr = p.communicate(stdin)

    if p.returncode != 0:
        raise RuntimeError(stderr)


def kinit_password(principal, password, ccache_name, config):
    """
    Perform kinit using principal/password.

    It uses the specified config file to kinit and stores the TGT
    in ccache_name.
    """
    args = ["/usr/bin/kinit", principal, '-c', ccache_name]
    old_config = os.environ.get('KRB5_CONFIG')
    os.environ['KRB5_CONFIG'] = config

    try:
        return run_cmd(args, stdin=password.encode())
    finally:
        if old_config is not None:
            os.environ['KRB5_CONFIG'] = old_config
        else:
            os.environ.pop('KRB5_CONFIG', None)


def kinit_keytab(principal, keytab, ccache_name, config):
    """
    Perform kinit using principal/keytab.

    It uses the specified config file to kinit and stores the TGT
    in ccache_name.
    """
    if gssapi is None:
        raise ImportError("gssapi is not available")

    old_config = os.environ.get('KRB5_CONFIG')
    os.environ['KRB5_CONFIG'] = config
    try:
        name = gssapi.Name(principal, gssapi.NameType.kerberos_principal)
        store = {'ccache': ccache_name,
                 'client_keytab': keytab}
        cred = gssapi.Credentials(name=name, store=store, usage='initiate')
        return cred
    finally:
        if old_config is not None:
            os.environ['KRB5_CONFIG'] = old_config
        else:
            os.environ.pop('KRB5_CONFIG', None)


KRB5CONF_TEMPLATE = """
[logging]
 default = FILE:/var/log/krb5libs.log
 kdc = FILE:/var/log/krb5kdc.log
 admin_server = FILE:/var/log/kadmind.log

[libdefaults]
 default_realm = {{ ipa_realm }}
 dns_lookup_realm = false
 dns_lookup_kdc = true
 rdns = false
 ticket_lifetime = {{ ipa_lifetime }}
 forwardable = true
 udp_preference_limit = 0
 default_ccache_name = KEYRING:persistent:%{uid}

[realms]
 {{ ipa_realm }} = {
  kdc = {{ ipa_server }}:88
  master_kdc = {{ ipa_server }}:88
  admin_server = {{ ipa_server }}:749
  default_domain = {{ ipa_domain }}
}

[domain_realm]
 .{{ ipa_domain }} = {{ ipa_realm }}
 {{ ipa_domain }} = {{ ipa_realm }}
"""


class ActionModule(ActionBase):

    def run(self, tmp=None, task_vars=None):
        """
        Handle credential cache transfer.

        ipa* commands can either provide a password or a keytab file
        in order to authenticate on the managed node with Kerberos.
        The module is using these credentials to obtain a TGT locally on the
        control node:
        - need to create a krb5.conf Kerberos client configuration that is
        using IPA server
        - set the environment variable KRB5_CONFIG to point to this conf file
        - set the environment variable KRB5CCNAME to use a specific cache
        - perform kinit on the control node
        This command creates the credential cache file
        - copy the credential cache file on the managed node

        Then the IPA commands can use this credential cache file.
        """
        if task_vars is None:
            task_vars = dict()

        result = super(ActionModule, self).run(tmp, task_vars)
        principal = self._task.args.get('principal', None)
        keytab = self._task.args.get('keytab', None)
        password = self._task.args.get('password', None)
        lifetime = self._task.args.get('lifetime', '1h')

        if (not keytab and not password):
            result['failed'] = True
            result['msg'] = "keytab or password is required"
            return result

        if not principal:
            result['failed'] = True
            result['msg'] = "principal is required"
            return result

        data = self._execute_module(module_name='ipaclient_get_facts',
                                    module_args=dict(), task_vars=task_vars)

        try:
            domain = data['ansible_facts']['ipa']['domain']
            realm = data['ansible_facts']['ipa']['realm']
        except KeyError:
            result['failed'] = True
            result['msg'] = "The host is not an IPA server"
            return result

        items = principal.split('@')
        if len(items) < 2:
            principal = str('%s@%s' % (principal, realm))

        # Locally create a temp directory to store krb5.conf and ccache
        local_temp_dir = tempfile.mkdtemp()
        krb5conf_name = os.path.join(local_temp_dir, 'krb5.conf')
        ccache_name = os.path.join(local_temp_dir, 'ccache')

        # Create the krb5.conf from the template
        template = Template(KRB5CONF_TEMPLATE)
        content = template.render(dict(
            ipa_server=task_vars['ansible_host'],
            ipa_domain=domain,
            ipa_realm=realm,
            ipa_lifetime=lifetime))

        with open(krb5conf_name, 'w') as f:
            f.write(content)

        if password:
            try:
                # perform kinit -c ccache_name -l 1h principal
                kinit_password(principal, password, ccache_name,
                               krb5conf_name)
            except Exception as e:
                result['failed'] = True
                result['msg'] = 'kinit %s with password failed: %s' % \
                    (principal, to_native(e))
                return result

        else:
            # Password not supplied, need to use the keytab file
            # Check if the source keytab exists
            try:
                keytab = self._find_needle('files', keytab)
            except AnsibleError as e:
                result['failed'] = True
                result['msg'] = to_native(e)
                return result
            # perform kinit -kt keytab
            try:
                kinit_keytab(principal, keytab, ccache_name, krb5conf_name)
            except Exception as e:
                result['failed'] = True
                result['msg'] = 'kinit %s with keytab %s failed: %s' % \
                    (principal, keytab, str(e))
                return result

        try:
            # Create the remote tmp dir
            tmp = self._make_tmp_path()
            tmp_ccache = self._connection._shell.join_path(
                tmp, os.path.basename(ccache_name))

            # Copy the ccache to the remote tmp dir
            self._transfer_file(ccache_name, tmp_ccache)
            self._fixup_perms2((tmp, tmp_ccache))

            new_module_args = self._task.args.copy()
            new_module_args.pop('password', None)
            new_module_args.pop('keytab', None)
            new_module_args.pop('lifetime', None)
            new_module_args.update(ccache=tmp_ccache)

            # Execute module
            result.update(self._execute_module(module_args=new_module_args,
                                               task_vars=task_vars))
            return result
        finally:
            # delete the local temp directory
            shutil.rmtree(local_temp_dir, ignore_errors=True)



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Authors:
#   Thomas Woerner <twoerner@redhat.com>
#
# Based on ipa-client-install code
#
# Copyright (C) 2017  Red Hat
# see file 'COPYING' for use and warranty information
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

ANSIBLE_METADATA = {
    'metadata_version': '1.0',
    'supported_by': 'community',
    'status': ['preview'],
}

DOCUMENTATION = '''
---
module: ipaclient_setup_nss
short description: Create IPA client NSS database
description: Create IPA NSS database
options:
  servers:
    description: Fully qualified name of IPA servers to enroll to
    required: no
  domain:
    description: Primary DNS domain of the IPA deployment
    required: no
  realm:
    description: Kerberos realm name of the IPA deployment
    required: no
  hostname:
    description: Fully qualified name of this host
    required: no
  basedn:
    description: The basedn of the IPA server (of the form dc=example,dc=com)
    required: no
  principal:
    description:
      User Principal allowed to promote replicas and join IPA realm
    required: yes
  subject_base:
    description: |
      The certificate subject base (default O=<realm-name>).
      RDNs are in LDAP order (most specific RDN first).
    required: no
  ca_enabled:
    description: Whether the Certificate Authority is enabled or not
    required: no
  mkhomedir:
    description: Create home directories for users on their first login
    required: yes
  on_master:
    description: Whether the configuration is done on the master or not
    required: yes
  dnsok:
    description: The installer dnsok setting
    required: yes
  enable_dns_updates:
    description: |
      Configures the machine to attempt dns updates when the ip address
      changes
    required: yes
  all_ip_addresses:
    description: |
      All routable IP addresses configured on any interface will be added
      to DNS
    required: yes
  ip_addresses:
    description: List of Master Server IP Addresses
    required: yes
  request_cert:
    description: Request certificate for the machine
    required: yes
  preserve_sssd:
    description: Preserve old SSSD configuration if possible
    required: yes
  no_ssh:
    description: Do not configure OpenSSH client
    required: yes
  no_sshd:
    description: Do not configure OpenSSH server
    required: yes
  no_sudo:
    description: Do not configure SSSD as data source for sudo
    required: yes
  fixed_primary:
    description: Configure sssd to use fixed server as primary IPA server
    required: yes
  permit:
    description: Disable access rules by default, permit all access
    required: yes
  no_krb5_offline_passwords:
    description:
      Configure SSSD not to store user password when the server is offline
    required: yes
  no_dns_sshfp:
    description: Do not automatically create DNS SSHFP records
    required: yes
author:
    - Thomas Woerner
'''

EXAMPLES = '''
- name: Create IPA client NSS database
  ipaclient_setup_nss:
    servers: ["server1.example.com","server2.example.com"]
    domain: example.com
    realm: EXAMPLE.COM
    basedn: dc=example,dc=com
    hostname: client1.example.com
    subject_base: O=EXAMPLE.COM
    principal: admin
    ca_enabled: yes
'''

RETURN = '''
'''

import os
import time
import inspect

from ansible.module_utils.basic import AnsibleModule
from ansible.module_utils.ansible_ipa_client import (
    setup_logging,
    options, sysrestore, paths, ansible_module_get_parsed_ip_addresses,
    api, errors, create_ipa_nssdb, ipautil, ScriptError, CLIENT_INSTALL_ERROR,
    get_certs_from_ldap, DN, certstore, x509, logger, certdb,
    CalledProcessError, tasks, client_dns, configure_certmonger, services,
    update_ssh_keys, save_state, configure_ldap_conf, configure_nslcd_conf,
    nosssd_files, configure_openldap_conf, hardcode_ldap_server
)


def main():
    module = AnsibleModule(
        argument_spec=dict(
            servers=dict(required=True, type='list'),
            domain=dict(required=True),
            realm=dict(required=True),
            hostname=dict(required=True),
            basedn=dict(required=True),
            principal=dict(required=False),
            subject_base=dict(required=True),
            ca_enabled=dict(required=True, type='bool'),
            mkhomedir=dict(required=False, type='bool'),
            on_master=dict(required=False, type='bool'),
            dnsok=dict(required=False, type='bool', default=False),

            enable_dns_updates=dict(required=False, type='bool'),
            all_ip_addresses=dict(required=False, type='bool', default=False),
            ip_addresses=dict(required=False, type='list', default=None),
            request_cert=dict(required=False, type='bool', default=False),
            preserve_sssd=dict(required=False, type='bool'),
            no_ssh=dict(required=False, type='bool'),
            no_sshd=dict(required=False, type='bool'),
            no_sudo=dict(required=False, type='bool'),
            fixed_primary=dict(required=False, type='bool'),
            permit=dict(required=False, type='bool'),
            no_krb5_offline_passwords=dict(required=False, type='bool'),
            no_dns_sshfp=dict(required=False, type='bool', default=False),
        ),
        supports_check_mode=True,
    )

    module._ansible_debug = True
    setup_logging()

    cli_server = module.params.get('servers')
    cli_realm = module.params.get('realm')
    hostname = module.params.get('hostname')
    cli_basedn = module.params.get('basedn')
    cli_domain = module.params.get('domain')
    options.principal = module.params.get('principal')
    subject_base = module.params.get('subject_base')
    ca_enabled = module.params.get('ca_enabled')
    options.mkhomedir = module.params.get('mkhomedir')
    options.on_master = module.params.get('on_master')
    dnsok = module.params.get('dnsok')

    fstore = sysrestore.FileStore(paths.IPA_CLIENT_SYSRESTORE)
    statestore = sysrestore.StateFile(paths.IPA_CLIENT_SYSRESTORE)

    os.environ['KRB5CCNAME'] = paths.IPA_DNS_CCACHE

    options.dns_updates = module.params.get('enable_dns_updates')
    options.all_ip_addresses = module.params.get('all_ip_addresses')
    options.ip_addresses = ansible_module_get_parsed_ip_addresses(module)
    options.request_cert = module.params.get('request_cert')
    options.hostname = hostname
    options.host_name = hostname
    options.preserve_sssd = module.params.get('preserve_sssd')
    options.no_ssh = module.params.get('no_ssh')
    options.conf_ssh = not options.no_ssh
    options.no_sshd = module.params.get('no_sshd')
    options.conf_sshd = not options.no_sshd
    options.no_sudo = module.params.get('no_sudo')
    options.conf_sudo = not options.no_sudo
    options.primary = module.params.get('fixed_primary')
    options.permit = module.params.get('permit')
    options.no_krb5_offline_passwords = module.params.get(
        'no_krb5_offline_passwords')
    options.krb5_offline_passwords = not options.no_krb5_offline_passwords
    options.no_dns_sshfp = module.params.get('no_dns_sshfp')
    options.create_sshfp = not options.no_dns_sshfp
    options.no_sssd = False
    options.sssd = not options.no_sssd
    options.no_ac = False

    CCACHE_FILE = paths.IPA_DNS_CCACHE

    api.bootstrap(context='cli_installer',
                  confdir=paths.ETC_IPA,
                  debug=False,
                  delegate=False)
    api.finalize()

    api.Backend.rpcclient.connect()
    try:
        api.Backend.rpcclient.forward('ping')
    except errors.KerberosError:
        # Cannot connect to the server due to Kerberos error, trying with
        # delegate=True
        api.Backend.rpcclient.disconnect()
        api.Backend.rpcclient.connect(delegate=True)
        api.Backend.rpcclient.forward('ping')

    ##########################################################################

    try:

        # Create IPA NSS database
        try:
            create_ipa_nssdb()
        except ipautil.CalledProcessError as e:
            raise ScriptError(
                "Failed to create IPA NSS database: %s" % e,
                rval=CLIENT_INSTALL_ERROR)

        # Get CA certificates from the certificate store
        try:
            ca_certs = get_certs_from_ldap(cli_server[0], cli_basedn,
                                           cli_realm, ca_enabled)
        except errors.NoCertificateError:
            if ca_enabled:
                ca_subject = DN(('CN', 'Certificate Authority'), subject_base)
            else:
                ca_subject = None
            ca_certs = certstore.make_compat_ca_certs(ca_certs, cli_realm,
                                                      ca_subject)
        ca_certs_trust = [(c, n,
                           certstore.key_policy_to_trust_flags(t, True, u))
                          for (c, n, t, u) in ca_certs]

        if hasattr(paths, "KDC_CA_BUNDLE_PEM"):
            x509.write_certificate_list(
                [c for c, n, t, u in ca_certs if t is not False],
                paths.KDC_CA_BUNDLE_PEM,
                # mode=0o644
            )
        if hasattr(paths, "CA_BUNDLE_PEM"):
            x509.write_certificate_list(
                [c for c, n, t, u in ca_certs if t is not False],
                paths.CA_BUNDLE_PEM,
                # mode=0o644
            )

        # Add the CA certificates to the IPA NSS database
        logger.debug("Adding CA certificates to the IPA NSS database.")
        ipa_db = certdb.NSSDatabase(paths.IPA_NSSDB_DIR)
        for cert, nickname, trust_flags in ca_certs_trust:
            try:
                ipa_db.add_cert(cert, nickname, trust_flags)
            except CalledProcessError:
                raise ScriptError(
                    "Failed to add %s to the IPA NSS database." % nickname,
                    rval=CLIENT_INSTALL_ERROR)

        # Add the CA certificates to the platform-dependant systemwide CA
        # store
        tasks.insert_ca_certs_into_systemwide_ca_store(ca_certs)

        if not options.on_master:
            client_dns(cli_server[0], hostname, options)
            configure_certmonger(fstore, subject_base, cli_realm, hostname,
                                 options, ca_enabled)

        if hasattr(paths, "SSH_CONFIG_DIR"):
            ssh_config_dir = paths.SSH_CONFIG_DIR
        else:
            ssh_config_dir = services.knownservices.sshd.get_config_dir()
        update_ssh_keys(hostname, ssh_config_dir, options.create_sshfp)

        try:
            os.remove(CCACHE_FILE)
        except Exception:
            pass

        argspec_save_state = inspect.getargspec(save_state)

        # Name Server Caching Daemon. Disable for SSSD, use otherwise
        # (if installed)
        nscd = services.knownservices.nscd
        if nscd.is_installed():
            if "statestore" in argspec_save_state.args:
                save_state(nscd, statestore)
            else:
                save_state(nscd)
            nscd_service_action = None
            try:
                if options.sssd:
                    nscd_service_action = 'stop'
                    nscd.stop()
                else:
                    nscd_service_action = 'restart'
                    nscd.restart()
            except Exception:
                logger.warning(
                    "Failed to %s the %s daemon",
                    nscd_service_action, nscd.service_name)
                if not options.sssd:
                    logger.warning(
                        "Caching of users/groups will not be available")

            try:
                if options.sssd:
                    nscd.disable()
                else:
                    nscd.enable()
            except Exception:
                if not options.sssd:
                    logger.warning(
                        "Failed to configure automatic startup of the %s "
                        "daemon",
                        nscd.service_name)
                    logger.info(
                        "Caching of users/groups will not be "
                        "available after reboot")
                else:
                    logger.warning(
                        "Failed to disable %s daemon. Disable it manually.",
                        nscd.service_name)

        else:
            # this is optional service, just log
            if not options.sssd:
                logger.info(
                    "%s daemon is not installed, skip configuration",
                    nscd.service_name)

        nslcd = services.knownservices.nslcd
        if nslcd.is_installed():
            if "statestore" in argspec_save_state.args:
                save_state(nslcd, statestore)
            else:
                save_state(nslcd)

        retcode, conf = (0, None)

        if not options.no_ac:
            # Modify nsswitch/pam stack
            argspec = inspect.getargspec(tasks.modify_nsswitch_pam_stack)
            if "sudo" in argspec.args:
                tasks.modify_nsswitch_pam_stack(
                    sssd=options.sssd,
                    mkhomedir=options.mkhomedir,
                    statestore=statestore,
                    sudo=options.conf_sudo
                )
            else:
                tasks.modify_nsswitch_pam_stack(
                    sssd=options.sssd,
                    mkhomedir=options.mkhomedir,
                    statestore=statestore
                )

            if hasattr(paths, "AUTHSELECT") and paths.AUTHSELECT is not None:
                # authselect is used
                # if mkhomedir, make sure oddjobd is enabled and started
                if options.mkhomedir:
                    oddjobd = services.service('oddjobd', api)
                    running = oddjobd.is_running()
                    enabled = oddjobd.is_enabled()
                    statestore.backup_state('oddjobd', 'running', running)
                    statestore.backup_state('oddjobd', 'enabled', enabled)
                    try:
                        if not enabled:
                            oddjobd.enable()
                        if not running:
                            oddjobd.start()
                    except Exception as e:
                        logger.critical("Unable to start oddjobd: %s", str(e))

            logger.info("%s enabled", "SSSD" if options.sssd else "LDAP")

            if options.sssd:
                sssd = services.service('sssd', api)
                try:
                    sssd.restart()
                except CalledProcessError:
                    logger.warning("SSSD service restart was unsuccessful.")

                try:
                    sssd.enable()
                except CalledProcessError as e:
                    logger.warning(
                        "Failed to enable automatic startup of the SSSD "
                        "daemon: %s", e)

            if not options.sssd:
                tasks.modify_pam_to_use_krb5(statestore)
                logger.info("Kerberos 5 enabled")

            # Update non-SSSD LDAP configuration after authconfig calls as it
            # would change its configuration otherways
            if not options.sssd:
                for configurer in [configure_ldap_conf, configure_nslcd_conf]:
                    (retcode, conf, filenames) = configurer(
                        fstore, cli_basedn, cli_realm,
                        cli_domain, cli_server, dnsok,
                        options, nosssd_files[configurer.__name__])
                    if retcode:
                        raise ScriptError(rval=CLIENT_INSTALL_ERROR)
                    if conf:
                        logger.info(
                            "%s configured using configuration file(s) %s",
                            conf, filenames)

            if configure_openldap_conf(fstore, cli_basedn, cli_server):
                logger.info("Configured /etc/openldap/ldap.conf")
            else:
                logger.info("Failed to configure /etc/openldap/ldap.conf")

            # Check that nss is working properly
            if not options.on_master:
                user = options.principal
                if user is None:
                    user = "admin@%s" % cli_domain
                    logger.info("Principal is not set when enrolling with OTP"
                                "; using principal '%s' for 'getent passwd'",
                                user)
                elif '@' not in user:
                    user = "%s@%s" % (user, cli_domain)
                n = 0
                found = False
                # Loop for up to 10 seconds to see if nss is working properly.
                # It can sometimes take a few seconds to connect to the remote
                # provider.
                # Particulary, SSSD might take longer than 6-8 seconds.
                if hasattr(paths, "GETENT"):
                    getent_cmd = paths.GETENT
                else:
                    getent_cmd = '/usr/bin/getent'
                while n < 10 and not found:
                    try:
                        ipautil.run([getent_cmd, "passwd", user])
                        found = True
                    except Exception:
                        time.sleep(1)
                        n = n + 1

                if not found:
                    logger.error("Unable to find '%s' user with 'getent "
                                 "passwd %s'!", user.split("@")[0], user)
                    if conf:
                        logger.info("Recognized configuration: %s", conf)
                    else:
                        logger.error(
                            "Unable to reliably detect "
                            "configuration. Check NSS setup manually.")

                    try:
                        hardcode_ldap_server(cli_server)
                    except Exception as e:
                        logger.error(
                            "Adding hardcoded server name to "
                            "/etc/ldap.conf failed: %s", str(e))

    except ScriptError as e:
        module.fail_json(msg=str(e))

    ##########################################################################

    module.exit_json(changed=True,
                     ca_enabled_ra=ca_enabled)


if __name__ == '__main__':
    main()



#!/usr/bin/env python

# Authors:
#   Sergio Oliveira Campos <seocam@redhat.com>
#
# Copyright (C) 2020 Red Hat
# see file 'COPYING' for use and warranty information
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import pytest
import subprocess
import tempfile
import testinfra

from unittest import TestCase


SCRIPT_DIR = os.path.dirname(os.path.abspath(__file__))


def is_docker_env():
    if os.getenv("RUN_TESTS_IN_DOCKER", "0") == "0":
        return False
    return True


def get_ssh_password():
    return os.getenv("IPA_SSH_PASSWORD")


def get_server_host():
    return os.getenv("IPA_SERVER_HOST")


def get_inventory_content():
    """Create the content of an inventory file for a test run."""
    ipa_server_host = get_server_host()

    if is_docker_env():
        ipa_server_host += " ansible_connection=docker"

    sshpass = get_ssh_password()
    if sshpass:
        ipa_server_host += " ansible_ssh_pass=%s" % sshpass

    lines = [
        "[ipaserver]",
        ipa_server_host,
        "[ipaserver:vars]",
        "ipaserver_domain=test.local",
        "ipaserver_realm=TEST.LOCAL",
    ]
    return "\n".join(lines).encode("utf8")


def get_test_name_from_playbook_path(playbook):
    """
    Create a test name based of a playbook path.

    For example:
        Input: /home/johndoe/ansible-freeipa/tests/dnszone/test_dnszone_mod.yml
        Output: dnszone_test_dnszone_mod
    """
    playbook_abspath = os.path.abspath(playbook)
    playbook_rel_to_tests_dir = playbook_abspath.replace(SCRIPT_DIR, "")
    playbook_slug = playbook_rel_to_tests_dir.strip("/").replace("/", "_")
    return os.path.splitext(playbook_slug)[0]


def write_logs(result, test_name):
    """Write logs of a ansible run logs to `test/logs/`."""
    log_dir = os.path.join(SCRIPT_DIR, "logs")
    if not os.path.exists(log_dir):
        os.makedirs(log_dir)

    # Write stdout log for test
    log_path = os.path.join(log_dir, "ansible_" + test_name + ".log")
    with open(log_path, "w") as log_file:
        log_file.write(result.stdout.decode("utf-8"))

    # Write stderr log for test
    error_log_path = os.path.join(log_dir, test_name + "-error.log")
    with open(error_log_path, "w") as log_file:
        log_file.write(result.stderr.decode("utf-8"))


def _run_playbook(playbook):
    """
    Create a inventory using a temporary file and run ansible using it.

    The logs of the run will be placed in `tests/logs/`.
    """
    with tempfile.NamedTemporaryFile() as inventory_file:
        inventory_file.write(get_inventory_content())
        inventory_file.flush()
        cmd = [
            "ansible-playbook",
            "-i",
            inventory_file.name,
            playbook,
        ]
        process = subprocess.run(
            cmd, cwd=SCRIPT_DIR, stdout=subprocess.PIPE, stderr=subprocess.PIPE
        )
    test_name = get_test_name_from_playbook_path(playbook)
    write_logs(process, test_name)

    return process


def run_playbook(playbook, allow_failures=False):
    """
    Run an Ansible playbook and assert the return code.

    Call ansible (using _run_playbook function) and assert the result of
    the execution.

    In case of failure the tail of the error message will be displayed
    as an assertion message.

    The full log of the execution will be available in the directory
    `tests/logs/`.
    """
    result = _run_playbook(playbook)

    if allow_failures:
        return result

    status_code_msg = "ansible-playbook return code: {}".format(
        result.returncode
    )
    assert_msg = "\n".join(
        [
            "",
            "-" * 30 + " Captured stdout " + "-" * 30,
            result.stdout.decode("utf8"),
            "-" * 30 + " Captured stderr " + "-" * 30,
            result.stderr.decode("utf8"),
            "-" * 30 + " Playbook Return Code " + "-" * 30,
            status_code_msg,
        ]
    )

    # Need to get the last bytes of msg otherwise Azure
    #   will cut it out.
    assert result.returncode == 0, assert_msg[-2500:]

    return result


def list_test_yaml(dir_path):
    """
    List the test playbooks inside a given directory.

    A test playbook is any file inside the directory which the name starts with
    `test_` and the extension is `.yml`.
    """
    yamls = []
    for root, dirs, files in os.walk(dir_path):
        for yaml_name in files:
            if yaml_name.startswith("test_") and yaml_name.endswith(".yml"):
                test_yaml_path = os.path.join(root, yaml_name)
                yamls.append(
                    {
                        "path": test_yaml_path,
                        "name": yaml_name.split(".")[0],
                    }
                )
    return yamls


def get_test_playbooks():
    """
    Get playbook tests grouped by first level directory.

    This function visits the first level of directories inside `tests/` and
    look for ansible playbooks on them.

    Returns a dict with the directories found in `tests/` as key and a
    list of test playbook files inside of it.

    A test playbook is any file inside the directory which the name starts with
    `test_` and the extension is `.yml`.
    """
    test_dirs = os.listdir(SCRIPT_DIR)
    groups = {}
    for test_group_dir in sorted(test_dirs):
        group_dir_path = os.path.join(SCRIPT_DIR, test_group_dir)
        if not os.path.isdir(group_dir_path):
            continue
        yamls = list_test_yaml(group_dir_path)
        if yamls:
            groups[test_group_dir] = yamls
    return groups


def kinit_admin(host, admin="admin", password="SomeADMINpassword"):
    return host.run_test("kinit " + admin + "<<< " + password)


def kdestroy(host):
    return host.run_test("kdestroy -A")


class AnsibleFreeIPATestCase(TestCase):
    def setUp(self):
        if is_docker_env():
            protocol = "docker://"
            user = ""
            ssh_identity_file = None
        else:
            protocol = "ssh://"

            password = get_ssh_password() or ""
            if password:
                password = ":" + password

            current_user = os.getenv("USER")
            ansible_user = os.getenv("ANSIBLE_REMOTE_USER", current_user)
            user = ansible_user + password + "@"
            ssh_identity_file = os.getenv("ANSIBLE_PRIVATE_KEY_FILE", None)

        host_connection_info = protocol + user + get_server_host()
        self.master = testinfra.get_host(
            host_connection_info, ssh_identity_file=ssh_identity_file,
        )

    def run_playbook(self, playbook, allow_failures=False):
        return run_playbook(playbook, allow_failures)

    def run_playbook_with_exp_msg(self, playbook, expected_msg):
        result = self.run_playbook(playbook, allow_failures=True)
        assert (
            expected_msg in result.stdout.decode("utf8")
            or
            expected_msg in result.stderr.decode("utf8")
        )

    def check_details(self, expected_output, cmd, extra_cmds=None):
        cmd = "ipa " + cmd
        if extra_cmds:
            cmd += " " + " ".join(extra_cmds)
        kinit_admin(self.master)
        res = self.master.run(cmd)
        if res.rc != 0:
            for output in expected_output:
                assert output in res.stderr
        else:
            for output in expected_output:
                assert output in res.stdout
        kdestroy(self.master)

    def check_notexists(self, members, cmd, extra_cmds=None):
        cmd = "ipa " + cmd
        if extra_cmds:
            cmd += " " + " ".join(extra_cmds)
        kinit_admin(self.master)
        res = self.master.run(cmd)
        for member in members:
            assert member not in res.stdout
        kdestroy(self.master)

    def mark_xfail_using_ansible_freeipa_version(self, version, reason):
        package = self.master.package("ansible-freeipa")

        if not package.is_installed:
            return

        if package.version == version:
            pytest.xfail(reason)



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Authors:
#   Thomas Woerner <twoerner@redhat.com>
#
# Based on ipa-replica-install code
#
# Copyright (C) 2018  Red Hat
# see file 'COPYING' for use and warranty information
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

ANSIBLE_METADATA = {
    'metadata_version': '1.0',
    'supported_by': 'community',
    'status': ['preview'],
}

DOCUMENTATION = '''
---
module: ipareplica_test
short description: IPA replica deployment tests
description: IPA replica deployment tests
options:
  ip_addresses:
    description: List of Master Server IP Addresses
    required: yes
  domain:
    description: Primary DNS domain of the IPA deployment
    required: yes
  servers:
    description: Fully qualified name of IPA servers to enroll to
    required: yes
  realm:
    description: Kerberos realm name of the IPA deployment
    required: yes
  hostname:
    description: Fully qualified name of this host
    required: yes
  ca_cert_files:
    description:
      List of files containing CA certificates for the service certificate
      files
    required: yes
  hidden_replica:
    description: Install a hidden replica
    required: yes
  skip_mem_check:
    description: Skip checking for minimum required memory
    required: yes
  setup_adtrust:
    description: Configure AD trust capability
    required: yes
  setup_ca:
    description: Configure a dogtag CA
    required: yes
  setup_kra:
    description: Configure a dogtag KRA
    required: yes
  setup_dns:
    description: Configure bind with our zone
    required: yes
  no_pkinit:
    description: Disable pkinit setup steps
    required: yes
  dirsrv_config_file:
    description:
      The path to LDIF file that will be used to modify configuration of
      dse.ldif during installation of the directory server instance
    required: yes
  dirsrv_cert_files:
    description:
      Files containing the Directory Server SSL certificate and private key
    required: yes
  http_cert_files:
    description:
      File containing the Apache Server SSL certificate and private key
    required: yes
  pkinit_cert_files:
    description:
      File containing the Kerberos KDC SSL certificate and private key
    required: yes
  no_ntp:
    description: Do not configure ntp
    required: yes
  ntp_servers:
    description: ntp servers to use
    required: yes
  ntp_pool:
    description: ntp server pool to use
    required: yes
  no_reverse:
    description: Do not create new reverse DNS zone
    required: yes
  auto_reverse:
    description: Create necessary reverse zones
    required: yes
  forwarders:
    description: Add DNS forwarders
    required: yes
  no_forwarders:
    description: Do not add any DNS forwarders, use root servers instead
    required: yes
  auto_forwarders:
    description: Use DNS forwarders configured in /etc/resolv.conf
    required: yes
  forward_policy:
    description: DNS forwarding policy for global forwarders
    required: yes
  no_dnssec_validation:
    description: Disable DNSSEC validation
    required: yes
author:
    - Thomas Woerner
'''

EXAMPLES = '''
'''

RETURN = '''
'''

import os
import inspect

from ansible.module_utils.basic import AnsibleModule
from ansible.module_utils.ansible_ipa_replica import (
    AnsibleModuleLog, setup_logging, options, installer, paths, sysrestore,
    ansible_module_get_parsed_ip_addresses, service,
    redirect_stdout, create_ipa_conf, ipautil,
    x509, validate_domain_name, common_check,
    IPA_PYTHON_VERSION
)


def main():
    ansible_module = AnsibleModule(
        argument_spec=dict(
            # basic
            # dm_password=dict(required=False, no_log=True),
            # password=dict(required=False, no_log=True),
            ip_addresses=dict(required=False, type='list', default=[]),
            domain=dict(required=False),
            servers=dict(required=False, type='list', default=[]),
            realm=dict(required=False),
            hostname=dict(required=False),
            ca_cert_files=dict(required=False, type='list', default=[]),
            hidden_replica=dict(required=False, type='bool', default=False),
            skip_mem_check=dict(required=False, type='bool', default=False),
            # server
            setup_adtrust=dict(required=False, type='bool', default=False),
            setup_ca=dict(required=False, type='bool'),
            setup_kra=dict(required=False, type='bool', default=False),
            setup_dns=dict(required=False, type='bool', default=False),
            no_pkinit=dict(required=False, type='bool', default=False),
            dirsrv_config_file=dict(required=False),
            # ssl certificate
            dirsrv_cert_files=dict(required=False, type='list', default=[]),
            http_cert_files=dict(required=False, type='list', default=[]),
            pkinit_cert_files=dict(required=False, type='list', default=[]),
            # client
            no_ntp=dict(required=False, type='bool', default=False),
            ntp_servers=dict(required=False, type='list', default=[]),
            ntp_pool=dict(required=False),
            # dns
            no_reverse=dict(required=False, type='bool', default=False),
            auto_reverse=dict(required=False, type='bool', default=False),
            forwarders=dict(required=False, type='list', default=[]),
            no_forwarders=dict(required=False, type='bool', default=False),
            auto_forwarders=dict(required=False, type='bool', default=False),
            forward_policy=dict(default=None, choices=['first', 'only']),
            no_dnssec_validation=dict(required=False, type='bool',
                                      default=False),
        ),
    )

    ansible_module._ansible_debug = True
    setup_logging()
    ansible_log = AnsibleModuleLog(ansible_module)

    # get parameters #

    # basic
    # options.dm_password = ansible_module.params.get('dm_password')
    # # options.password = ansible_module.params.get('password')
    # options.password = options.dm_password
    options.ip_addresses = ansible_module_get_parsed_ip_addresses(
        ansible_module)
    options.domain_name = ansible_module.params.get('domain')
    options.servers = ansible_module.params.get('servers')
    options.realm_name = ansible_module.params.get('realm')
    options.host_name = ansible_module.params.get('hostname')
    options.ca_cert_files = ansible_module.params.get('ca_cert_files')
    options.hidden_replica = ansible_module.params.get('hidden_replica')
    options.skip_mem_check = ansible_module.params.get('skip_mem_check')
    # server
    options.setup_adtrust = ansible_module.params.get('setup_adtrust')
    options.setup_ca = ansible_module.params.get('setup_ca')
    options.setup_kra = ansible_module.params.get('setup_kra')
    options.setup_dns = ansible_module.params.get('setup_dns')
    options.no_pkinit = ansible_module.params.get('no_pkinit')
    options.dirsrv_config_file = ansible_module.params.get(
        'dirsrv_config_file')
    # ssl certificate
    options.dirsrv_cert_files = ansible_module.params.get('dirsrv_cert_files')
    options.http_cert_files = ansible_module.params.get('http_cert_files')
    options.pkinit_cert_files = ansible_module.params.get('pkinit_cert_files')
    # client
    options.no_ntp = ansible_module.params.get('no_ntp')
    options.ntp_servers = ansible_module.params.get('ntp_servers')
    options.ntp_pool = ansible_module.params.get('ntp_pool')
    # dns
    options.no_reverse = ansible_module.params.get('no_reverse')
    options.auto_reverse = ansible_module.params.get('auto_reverse')
    options.forwarders = ansible_module.params.get('forwarders')
    options.no_forwarders = ansible_module.params.get('no_forwarders')
    options.auto_forwarders = ansible_module.params.get('auto_forwarders')
    options.forward_policy = ansible_module.params.get('forward_policy')
    options.no_dnssec_validation = ansible_module.params.get(
        'no_dnssec_validation')

    ##########################################################################
    # replica init ###########################################################
    ##########################################################################

    if installer.servers:
        installer.server = installer.servers[0]
    else:
        installer.server = None
    # TODO: Kills ipa-client-install
    # if installer.replica_file is None:
    #     installer.password = installer.admin_password
    # else:
    #     installer.password = installer.dm_password

    # installer._ccache = os.environ.get('KRB5CCNAME')

    # If not defined, set domain from server name
    if installer.domain_name is None and installer.server is not None:
        installer.domain_name = installer.server[installer.server.find(".")+1:]
    # If not defined, set realm from domain name
    if installer.realm_name is None and installer.domain_name is not None:
        installer.realm_name = installer.domain_name.upper()

    ##########################################################################
    # other checks ###########################################################
    ##########################################################################

    # version specific tests #

    # if options.setup_adtrust and not adtrust_imported:
    #    # if "adtrust" not in options._allow_missing:
    #    ansible_module.fail_json(msg="adtrust can not be imported")
    #    # else:
    #    #  options.setup_adtrust = False
    #    #  ansible_module.warn(msg="adtrust is not supported, disabling")

    # if options.setup_kra and not kra_imported:
    #    # if "kra" not in options._allow_missing:
    #    ansible_module.fail_json(msg="kra can not be imported")
    #    # else:
    #    #  options.setup_kra = False
    #    #  ansible_module.warn(msg="kra is not supported, disabling")

    if options.hidden_replica and not hasattr(service, "hide_services"):
        ansible_module.fail_json(
            msg="Hidden replica is not supported in this version.")

    # We need to point to the master in ipa default conf when certmonger
    # asks for HTTP certificate in newer ipa versions. In these versions
    # create_ipa_conf has the additional master argument.
    change_master_for_certmonger = False
    argspec = inspect.getargspec(create_ipa_conf)
    if "master" in argspec.args:
        change_master_for_certmonger = True

    # From ipa installer classes

    # pkinit is not supported on DL0, don't allow related options
    if installer.replica_file is not None:
        ansible_module.fail_json(
            msg="Replica installation using a replica file is not supported")

    # If any of the key file options are selected, all are required.
    cert_file_req = (installer.dirsrv_cert_files, installer.http_cert_files)
    cert_file_opt = (installer.pkinit_cert_files,)
    if not installer.no_pkinit:
        cert_file_req += cert_file_opt
    if installer.no_pkinit and installer.pkinit_cert_files:
        ansible_module.fail_json(
            msg="--no-pkinit and --pkinit-cert-file cannot be specified "
            "together")
    if any(cert_file_req + cert_file_opt) and not all(cert_file_req):
        ansible_module.fail_json(
            msg="--dirsrv-cert-file, --http-cert-file, and --pkinit-cert-file "
            "or --no-pkinit are required if any key file options are used.")

    if not installer.setup_dns:
        if installer.forwarders:
            ansible_module.fail_json(
                msg="You cannot specify a --forwarder option without the "
                "--setup-dns option")
        if installer.auto_forwarders:
            ansible_module.fail_json(
                msg="You cannot specify a --auto-forwarders option without "
                "the --setup-dns option")
        if installer.no_forwarders:
            ansible_module.fail_json(
                msg="You cannot specify a --no-forwarders option without the "
                "--setup-dns option")
        if installer.forward_policy:
            ansible_module.fail_json(
                msg="You cannot specify a --forward-policy option without the "
                "--setup-dns option")
        if installer.reverse_zones:
            ansible_module.fail_json(
                msg="You cannot specify a --reverse-zone option without the "
                "--setup-dns option")
        if installer.auto_reverse:
            ansible_module.fail_json(
                msg="You cannot specify a --auto-reverse option without the "
                "--setup-dns option")
        if installer.no_reverse:
            ansible_module.fail_json(
                msg="You cannot specify a --no-reverse option without the "
                "--setup-dns option")
        if installer.no_dnssec_validation:
            ansible_module.fail_json(
                msg="You cannot specify a --no-dnssec-validation option "
                "without the --setup-dns option")
    elif installer.forwarders and installer.no_forwarders:
        ansible_module.fail_json(
            msg="You cannot specify a --forwarder option together with "
            "--no-forwarders")
    elif installer.auto_forwarders and installer.no_forwarders:
        ansible_module.fail_json(
            msg="You cannot specify a --auto-forwarders option together with "
            "--no-forwarders")
    elif installer.reverse_zones and installer.no_reverse:
        ansible_module.fail_json(
            msg="You cannot specify a --reverse-zone option together with "
            "--no-reverse")
    elif installer.auto_reverse and installer.no_reverse:
        ansible_module.fail_json(
            msg="You cannot specify a --auto-reverse option together with "
            "--no-reverse")

    # replica installers
    if installer.servers and not installer.domain_name:
        ansible_module.fail_json(
            msg="The --server option cannot be used without providing "
            "domain via the --domain option")

    if installer.setup_dns:
        if (not installer.forwarders and
                not installer.no_forwarders and
                not installer.auto_forwarders):
            ansible_module.fail_json(
                msg="You must specify at least one of --forwarder, "
                "--auto-forwarders, or --no-forwarders options")

    if installer.dirsrv_config_file is not None and \
       not os.path.exists(installer.dirsrv_config_file):
        ansible_module.fail_json(
            msg="File %s does not exist." % installer.dirsrv_config_file)

    if installer.ca_cert_files is not None:
        if not isinstance(installer.ca_cert_files, list):
            ansible_module.fail_json(
                msg="Expected list, got {!r}".format(installer.ca_cert_files))
        for cert in installer.ca_cert_files:
            if not os.path.exists(cert):
                ansible_module.fail_json(msg="'%s' does not exist" % cert)
            if not os.path.isfile(cert):
                ansible_module.fail_json(msg="'%s' is not a file" % cert)
            if not os.path.isabs(cert):
                ansible_module.fail_json(
                    msg="'%s' is not an absolute file path" % cert)

            try:
                x509.load_certificate_from_file(cert)
            except Exception:
                ansible_module.fail_json(
                    msg="'%s' is not a valid certificate file" % cert)

    if installer.ip_addresses is not None:
        for value in installer.ip_addresses:
            try:
                ipautil.CheckedIPAddress(value)
            except Exception as e:
                ansible_module.fail_json(
                    msg="invalid IP address {0}: {1}".format(
                        value, e))

    if installer.domain_name is not None:
        validate_domain_name(installer.domain_name)

    ##########################################################################
    # replica promote_check excerpts #########################################
    ##########################################################################

    # check selinux status, http and DS ports, NTP conflicting services
    try:
        with redirect_stdout(ansible_log):
            argspec = inspect.getargspec(common_check)
            if "skip_mem_check" in argspec.args:
                common_check(options.no_ntp, options.skip_mem_check,
                             options.setup_ca)
            else:
                common_check(options.no_ntp)
    except Exception as msg:  # ScriptError as msg:
        _msg = str(msg)
        if "server is already configured" in _msg:
            ansible_module.exit_json(changed=False,
                                     server_already_configured=True)
        else:
            ansible_module.fail_json(msg=_msg)

    # TODO: Check ntp_servers and ntp_pool

    # client enrolled?

    client_fstore = sysrestore.FileStore(paths.IPA_CLIENT_SYSRESTORE)
    client_enrolled = client_fstore.has_files()

    if not client_enrolled:
        # # One-step replica installation
        # if options.dm_password and options.password:
        #    ansible_module.fail_json(
        #        msg="--password and --admin-password options are "
        #        "mutually exclusive")
        pass
    else:
        # The NTP configuration can not be touched on pre-installed client:
        if options.no_ntp or options.ntp_servers or options.ntp_pool:
            ansible_module.fail_json(
                msg="NTP configuration cannot be updated during promotion")

    # done #

    ansible_module.exit_json(
        changed=False,
        ipa_python_version=IPA_PYTHON_VERSION,
        # basic
        domain=options.domain_name,
        realm=options.realm_name,
        hostname=options.host_name,
        # server
        setup_adtrust=options.setup_adtrust,
        setup_kra=options.setup_kra,
        server=options.server,
        # additional
        client_enrolled=client_enrolled,
        change_master_for_certmonger=change_master_for_certmonger,
    )


if __name__ == '__main__':
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Authors:
#   Thomas Woerner <twoerner@redhat.com>
#
# Copyright (C) 2019 Red Hat
# see file 'COPYING' for use and warranty information
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

ANSIBLE_METADATA = {
    "metadata_version": "1.0",
    "supported_by": "community",
    "status": ["preview"],
}

DOCUMENTATION = """
---
module: ipatopologysegment
short description: Manage FreeIPA topology segments
description: Manage FreeIPA topology segments
options:
  ipaadmin_principal:
    description: The admin principal
    default: admin
  ipaadmin_password:
    description: The admin password
    required: false
  suffix:
    description: Topology suffix
    required: true
    choices: ["domain", "ca", "domain+ca"]
  name:
    description: Topology segment name, unique identifier.
    required: false
    aliases: ["cn"]
  left:
    description: Left replication node - an IPA server
    aliases: ["leftnode"]
  right:
    description: Right replication node - an IPA server
    aliases: ["rightnode"]
  direction:
    description: The direction a segment will be reinitialized
    required: false
    choices: ["left-to-right", "right-to-left"]
  state:
    description: State to ensure
    default: present
    choices: ["present", "absent", "enabled", "disabled", "reinitialized",
              "checked" ]
author:
    - Thomas Woerner
"""

EXAMPLES = """
- ipatopologysegment:
    suffix: domain
    left: ipaserver.test.local
    right: ipareplica1.test.local
    state: present

- ipatopologysegment:
    suffix: domain
    name: ipaserver.test.local-to-replica1.test.local
    state: absent

- ipatopologysegment:
    suffix: domain
    left: ipaserver.test.local
    right: ipareplica1.test.local
    state: absent

- ipatopologysegment:
    suffix: ca
    name: ipaserver.test.local-to-replica1.test.local
    direction: left-to-right
    state: reinitialized

- ipatopologysegment:
    suffix: domain+ca
    left: ipaserver.test.local
    right: ipareplica1.test.local
    state: absent

- ipatopologysegment:
    suffix: domain+ca
    left: ipaserver.test.local
    right: ipareplica1.test.local
    state: checked
"""

RETURN = """
found:
  description: List of found segments
  returned: if state is checked
  type: list
not-found:
  description: List of not found segments
  returned: if state is checked
  type: list
"""

from ansible.module_utils.basic import AnsibleModule
from ansible.module_utils._text import to_text
from ansible.module_utils.ansible_freeipa_module import temp_kinit, \
    temp_kdestroy, valid_creds, api_connect, api_command


def find_left_right(module, suffix, left, right):
    _args = {
        "iparepltoposegmentleftnode": to_text(left),
        "iparepltoposegmentrightnode": to_text(right),
    }
    _result = api_command(module, "topologysegment_find",
                          to_text(suffix), _args)
    if len(_result["result"]) > 1:
        module.fail_json(
            msg="Combination of left node '%s' and right node '%s' is "
            "not unique for suffix '%s'" % (left, right, suffix))
    elif len(_result["result"]) == 1:
        return _result["result"][0]
    else:
        return None


def find_cn(module, suffix, name):
    _args = {
        "cn": to_text(name),
    }
    _result = api_command(module, "topologysegment_find",
                          to_text(suffix), _args)
    if len(_result["result"]) > 1:
        module.fail_json(
            msg="CN '%s' is not unique for suffix '%s'" % (name, suffix))
    elif len(_result["result"]) == 1:
        return _result["result"][0]
    else:
        return None


def find_left_right_cn(module, suffix, left, right, name):
    if left is not None and right is not None:
        left_right = find_left_right(module, suffix, left, right)
        if left_right is not None:
            if name is not None and \
               left_right["cn"][0] != to_text(name):
                module.fail_json(
                    msg="Left and right nodes do not match "
                    "given name name (cn) '%s'" % name)
            return left_right
        # else: Nothing to change
    elif name is not None:
        cn = find_cn(module, suffix, name)
        if cn is not None:
            return cn
        # else: Nothing to change
    else:
        module.fail_json(
            msg="Either left and right or name need to be set.")
    return None


def main():
    ansible_module = AnsibleModule(
        argument_spec=dict(
            ipaadmin_principal=dict(type="str", default="admin"),
            ipaadmin_password=dict(type="str", required=False, no_log=True),
            suffix=dict(choices=["domain", "ca", "domain+ca"], required=True),
            name=dict(type="str", aliases=["cn"], default=None),
            left=dict(type="str", aliases=["leftnode"], default=None),
            right=dict(type="str", aliases=["rightnode"], default=None),
            direction=dict(type="str", default=None,
                           choices=["left-to-right", "right-to-left"]),
            state=dict(type="str", default="present",
                       choices=["present", "absent", "enabled", "disabled",
                                "reinitialized", "checked"]),
        ),
        supports_check_mode=True,
    )

    ansible_module._ansible_debug = True

    # Get parameters

    ipaadmin_principal = ansible_module.params.get("ipaadmin_principal")
    ipaadmin_password = ansible_module.params.get("ipaadmin_password")
    suffixes = ansible_module.params.get("suffix")
    name = ansible_module.params.get("name")
    left = ansible_module.params.get("left")
    right = ansible_module.params.get("right")
    direction = ansible_module.params.get("direction")
    state = ansible_module.params.get("state")

    # Check parameters

    if state != "reinitialized" and direction is not None:
        ansible_module.fail_json(
            msg="Direction is not supported in this mode.")

    # Init

    changed = False
    exit_args = {}
    ccache_dir = None
    ccache_name = None
    try:
        if not valid_creds(ansible_module, ipaadmin_principal):
            ccache_dir, ccache_name = temp_kinit(ipaadmin_principal,
                                                 ipaadmin_password)
        api_connect()

        commands = []

        for suffix in suffixes.split("+"):
            # Create command
            if state in ["present", "enabled"]:
                # Make sure topology segment exists

                if left is None or right is None:
                    ansible_module.fail_json(
                        msg="Left and right need to be set.")
                args = {
                    "iparepltoposegmentleftnode": to_text(left),
                    "iparepltoposegmentrightnode": to_text(right),
                }
                if name is not None:
                    args["cn"] = to_text(name)

                res_left_right = find_left_right(ansible_module, suffix,
                                                 left, right)
                if res_left_right is not None:
                    if name is not None and \
                       res_left_right["cn"][0] != to_text(name):
                        ansible_module.fail_json(
                            msg="Left and right nodes already used with "
                            "different name (cn) '%s'" % res_left_right["cn"])

                    # Left and right nodes and also the name can not be
                    # changed
                    for key in ["iparepltoposegmentleftnode",
                                "iparepltoposegmentrightnode"]:
                        if key in args:
                            del args[key]
                    if len(args) > 1:
                        # cn needs to be in args always
                        commands.append(["topologysegment_mod", args, suffix])
                    # else: Nothing to change
                else:
                    if name is None:
                        args["cn"] = to_text("%s-to-%s" % (left, right))
                    commands.append(["topologysegment_add", args, suffix])

            elif state in ["absent", "disabled"]:
                # Make sure topology segment does not exist

                res_find = find_left_right_cn(ansible_module, suffix,
                                              left, right, name)
                if res_find is not None:
                    # Found either given name or found name from left and right
                    # node
                    args = {
                        "cn": res_find["cn"][0]
                    }
                    commands.append(["topologysegment_del", args, suffix])

            elif state == "checked":
                # Check if topology segment does exists

                res_find = find_left_right_cn(ansible_module, suffix,
                                              left, right, name)
                if res_find is not None:
                    # Found either given name or found name from left and right
                    # node
                    exit_args.setdefault("found", []).append(suffix)
                else:
                    # Not found
                    exit_args.setdefault("not-found", []).append(suffix)

            elif state == "reinitialized":
                # Reinitialize segment

                if direction not in ["left-to-right", "right-to-left"]:
                    ansible_module.fail_json(msg="Unknown direction '%s'" %
                                             direction)

                res_find = find_left_right_cn(ansible_module, suffix,
                                              left, right, name)
                if res_find is not None:
                    # Found either given name or found name from left and right
                    # node
                    args = {
                        "cn": res_find["cn"][0]
                    }
                    if direction == "left-to-right":
                        args["left"] = True
                    elif direction == "right-to-left":
                        args["right"] = True

                    commands.append(["topologysegment_reinitialize", args,
                                     suffix])
                else:
                    params = []
                    if name is not None:
                        params.append("name=%s" % name)
                    if left is not None:
                        params.append("left=%s" % left)
                    if right is not None:
                        params.append("right=%s" % right)
                    ansible_module.fail_json(
                        msg="No entry '%s' for suffix '%s'" %
                        (",".join(params), suffix))

            else:
                ansible_module.fail_json(msg="Unkown state '%s'" % state)

        # Execute command

        for command, args, _suffix in commands:
            api_command(ansible_module, command, to_text(_suffix), args)
            changed = True

    except Exception as e:
        ansible_module.fail_json(msg=str(e))

    finally:
        temp_kdestroy(ccache_dir, ccache_name)

    # Done

    ansible_module.exit_json(changed=changed, **exit_args)


if __name__ == "__main__":
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Authors:
#   Thomas Woerner <twoerner@redhat.com>
#
# Based on ipa-replica-install code
#
# Copyright (C) 2018  Red Hat
# see file 'COPYING' for use and warranty information
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

__all__ = ["contextlib", "dnsexception", "dnsresolver", "dnsreversename",
           "parse_version", "IPAChangeConf",
           "certstore", "sysrestore", "ipa_generate_password", "kinit_keytab",
           "IPA_CA_TRUST_FLAGS", "EXTERNAL_CA_TRUST_FLAGS", "DN",
           "ScriptError", "services", "tasks", "constants", "errors", "rpc",
           "x509", "validate_domain_name",
           "no_matching_interface_for_ip_address_warning",
           "configure_krb5_conf", "purge_host_keytab", "adtrust",
           "bindinstance", "ca", "certs", "dns", "httpinstance", "kra",
           "otpdinstance", "custodiainstance", "service", "upgradeinstance",
           "find_providing_servers", "find_providing_server", "load_pkcs12",
           "is_ipa_configured", "ReplicationManager", "replica_conn_check",
           "install_replica_ds", "install_krb", "install_ca_cert",
           "install_http", "install_dns_records", "create_ipa_conf",
           "check_dirsrv", "check_dns_resolution", "configure_certmonger",
           "remove_replica_info_dir", "preserve_enrollment_state",
           "uninstall_client", "promote_sssd", "promote_openldap_conf",
           "rpc_client", "check_remote_fips_mode", "check_remote_version",
           "common_check", "current_domain_level",
           "check_domain_level_is_supported", "promotion_check_ipa_domain",
           "SSSDConfig", "CalledProcessError", "timeconf", "ntpinstance",
           "dnsname", "kernel_keyring", "krbinstance"]

import sys
import logging
from contextlib import contextmanager as contextlib_contextmanager


from ipapython.version import NUM_VERSION, VERSION

if NUM_VERSION < 30201:
    # See ipapython/version.py
    IPA_MAJOR, IPA_MINOR, IPA_RELEASE = [int(x) for x in VERSION.split(".", 2)]
    IPA_PYTHON_VERSION = IPA_MAJOR*10000 + IPA_MINOR*100 + IPA_RELEASE
else:
    IPA_PYTHON_VERSION = NUM_VERSION


if NUM_VERSION >= 40600:
    # IPA version >= 4.6

    import contextlib

    import dns.exception as dnsexception
    import dns.name as dnsname
    import dns.resolver as dnsresolver
    import dns.reversename as dnsreversename

    from pkg_resources import parse_version

    from ipaclient.install.ipachangeconf import IPAChangeConf
    from ipalib.install import certstore, sysrestore
    from ipapython.ipautil import ipa_generate_password
    from ipalib.install.kinit import kinit_keytab
    from ipapython import ipaldap, ipautil, kernel_keyring
    from ipapython.certdb import IPA_CA_TRUST_FLAGS, EXTERNAL_CA_TRUST_FLAGS
    from ipapython.dn import DN
    from ipapython.admintool import ScriptError
    from ipapython.ipa_log_manager import standard_logging_setup
    from ipaplatform import services
    from ipaplatform.tasks import tasks
    from ipaplatform.paths import paths
    from ipalib import api, constants, create_api, errors, rpc, x509
    from ipalib.config import Env
    from ipalib.util import (
        validate_domain_name,
        no_matching_interface_for_ip_address_warning)
    from ipaclient.install.client import configure_krb5_conf, purge_host_keytab
    from ipaserver.install import (
        adtrust, bindinstance, ca, certs, dns, dsinstance, httpinstance,
        installutils, kra, krbinstance,
        otpdinstance, custodiainstance, service, upgradeinstance)
    try:
        from ipaserver.masters import (
            find_providing_servers, find_providing_server)
    except ImportError:
        from ipaserver.install.service import (
            find_providing_servers, find_providing_server)
    from ipaserver.install.installutils import (
        ReplicaConfig, load_pkcs12)
    try:
        from ipalib.facts import is_ipa_configured
    except ImportError:
        from ipaserver.install.installutils import is_ipa_configured
    from ipaserver.install.replication import (
        ReplicationManager, replica_conn_check)
    from ipaserver.install.server.replicainstall import (
        make_pkcs12_info, install_replica_ds, install_krb, install_ca_cert,
        install_http, install_dns_records, create_ipa_conf, check_dirsrv,
        check_dns_resolution, configure_certmonger, remove_replica_info_dir,
        # common_cleanup,
        preserve_enrollment_state, uninstall_client,
        promote_sssd, promote_openldap_conf, rpc_client,
        check_remote_fips_mode, check_remote_version, common_check,
        current_domain_level, check_domain_level_is_supported,
        # enroll_dl0_replica,
        # ensure_enrolled,
        promotion_check_ipa_domain
    )
    import SSSDConfig
    from subprocess import CalledProcessError

    try:
        from ipaclient.install import timeconf
        time_service = "chronyd"
        ntpinstance = None
    except ImportError:
        try:
            from ipaclient.install import ntpconf as timeconf
        except ImportError:
            from ipaclient import ntpconf as timeconf
        from ipaserver.install import ntpinstance
        time_service = "ntpd"


else:
    # IPA version < 4.6

    raise Exception("freeipa version '%s' is too old" % VERSION)


logger = logging.getLogger("ipa-server-install")


def setup_logging():
    # logger.setLevel(logging.DEBUG)
    standard_logging_setup(
        paths.IPAREPLICA_INSTALL_LOG, verbose=False, debug=False,
        filemode='a', console_format='%(message)s')


@contextlib_contextmanager
def redirect_stdout(f):
    sys.stdout = f
    try:
        yield f
    finally:
        sys.stdout = sys.__stdout__


class AnsibleModuleLog():
    def __init__(self, module):
        self.module = module
        _ansible_module_log = self

        class AnsibleLoggingHandler(logging.Handler):
            def emit(self, record):
                _ansible_module_log.write(self.format(record))

        self.logging_handler = AnsibleLoggingHandler()
        logger.setLevel(logging.DEBUG)
        logger.root.addHandler(self.logging_handler)

    def close(self):
        self.flush()

    def flush(self):
        pass

    def log(self, msg):
        # self.write(msg+"\n")
        self.write(msg)

    def debug(self, msg):
        self.module.debug(msg)

    def info(self, msg):
        self.module.debug(msg)

    def write(self, msg):
        self.module.debug(msg)
        # self.module.warn(msg)


class installer_obj(object):
    def __init__(self):
        # CompatServerReplicaInstall
        self.ca_cert_files = None
        self.all_ip_addresses = False
        self.no_wait_for_dns = True
        self.nisdomain = None
        self.no_nisdomain = False
        self.no_sudo = False
        self.request_cert = False
        self.ca_file = None
        self.zonemgr = None
        self.replica_file = None
        # ServerReplicaInstall
        self.subject_base = None
        self.ca_subject = None
        # others
        self._ccache = None
        self.password = None
        self.reverse_zones = []
        # def _is_promote(self):
        #     return self.replica_file is None
        # self.skip_conncheck = False
        self._replica_install = False
        # self.dnssec_master = False # future unknown
        # self.disable_dnssec_master = False # future unknown
        # self.domainlevel = MAX_DOMAIN_LEVEL # deprecated
        # self.domain_level = self.domainlevel # deprecated
        self.interactive = False
        self.unattended = not self.interactive
        # self.promote = self.replica_file is None
        self.promote = True
        self.skip_schema_check = None

    # def __getattribute__(self, attr):
    #     value = super(installer_obj, self).__getattribute__(attr)
    #     if not attr.startswith("--") and not attr.endswith("--"):
    #         logger.debug(
    #             "  <-- Accessing installer.%s (%s)" % (attr, repr(value)))
    #     return value

    def __getattr__(self, attr):
        logger.info("  --> ADDING missing installer.%s", attr)
        setattr(self, attr, None)
        return getattr(self, attr)

    # def __setattr__(self, attr, value):
    #    logger.debug("  --> Setting installer.%s to %s" % (attr, repr(value)))
    #    return super(installer_obj, self).__setattr__(attr, value)

    def knobs(self):
        for name in self.__dict__:
            yield self, name


installer = installer_obj()
options = installer

# DNSInstallInterface
options.dnssec_master = False
options.disable_dnssec_master = False
options.kasp_db_file = None
options.force = False

# ServerMasterInstall
options.add_sids = False
options.add_agents = False

# ServerReplicaInstall
options.subject_base = None
options.ca_subject = None


def gen_env_boostrap_finalize_core(etc_ipa, default_config):
    env = Env()
    # env._bootstrap(context='installer', confdir=paths.ETC_IPA, log=None)
    # env._finalize_core(**dict(constants.DEFAULT_CONFIG))
    env._bootstrap(context='installer', confdir=etc_ipa, log=None)
    env._finalize_core(**dict(default_config))
    return env


def api_bootstrap_finalize(env):
    # pylint: disable=no-member
    xmlrpc_uri = 'https://{}/ipa/xml'.format(ipautil.format_netloc(env.host))
    api.bootstrap(in_server=True,
                  context='installer',
                  confdir=paths.ETC_IPA,
                  ldap_uri=installutils.realm_to_ldapi_uri(env.realm),
                  xmlrpc_uri=xmlrpc_uri)
    # pylint: enable=no-member
    api.finalize()


def gen_ReplicaConfig():
    class ExtendedReplicaConfig(ReplicaConfig):
        def __init__(self, top_dir=None):
            super(ExtendedReplicaConfig, self).__init__(top_dir)

        # def __getattribute__(self, attr):
        #    value = super(ExtendedReplicaConfig, self).__getattribute__(attr)
        #    if attr not in ["__dict__", "knobs"]:
        #        logger.debug("  <== Accessing config.%s (%s)" %
        #                     (attr, repr(value)))
        #    return value

        def __getattr__(self, attr):
            logger.info("  ==> ADDING missing config.%s", attr)
            setattr(self, attr, None)
            return getattr(self, attr)

        # def __setattr__(self, attr, value):
        #   logger.debug("  ==> Setting config.%s to %s" % (attr, repr(value)))
        #   return super(ExtendedReplicaConfig, self).__setattr__(attr, value)

        def knobs(self):
            for name in self.__dict__:
                yield self, name

    # config = ReplicaConfig()
    config = ExtendedReplicaConfig()
    config.realm_name = api.env.realm
    config.host_name = api.env.host
    config.domain_name = api.env.domain
    config.master_host_name = api.env.server
    config.ca_host_name = api.env.ca_host
    config.kra_host_name = config.ca_host_name
    config.ca_ds_port = 389
    config.setup_ca = options.setup_ca
    config.setup_kra = options.setup_kra
    config.dir = options._top_dir
    config.basedn = api.env.basedn
    # config.subject_base = options.subject_base

    return config


def replica_ds_init_info(ansible_log,
                         config, options, ca_is_configured, remote_api,
                         ds_ca_subject, ca_file,
                         promote=False, pkcs12_info=None):

    dsinstance.check_ports()

    # if we have a pkcs12 file, create the cert db from
    # that. Otherwise the ds setup will create the CA
    # cert
    if pkcs12_info is None:
        pkcs12_info = make_pkcs12_info(config.dir, "dscert.p12",
                                       "dirsrv_pin.txt")

    # during replica install, this gets invoked before local DS is
    # available, so use the remote api.
    # if ca_is_configured:
    #     ca_subject = ca.lookup_ca_subject(_api, config.subject_base)
    # else:
    #     ca_subject = installutils.default_ca_subject_dn(config.subject_base)
    ca_subject = ds_ca_subject

    ds = dsinstance.DsInstance(
        config_ldif=options.dirsrv_config_file)
    ds.set_output(ansible_log)

    # Source: ipaserver/install/dsinstance.py

    # idstart and idmax are configured so that the range is seen as
    # depleted by the DNA plugin and the replica will go and get a
    # new range from the master.
    # This way all servers use the initially defined range by default.
    idstart = 1101
    idmax = 1100

    with redirect_stdout(ansible_log):
        ds.init_info(
            realm_name=config.realm_name,
            fqdn=config.host_name,
            domain_name=config.domain_name,
            dm_password=config.dirman_password,
            subject_base=config.subject_base,
            ca_subject=ca_subject,
            idstart=idstart,
            idmax=idmax,
            pkcs12_info=pkcs12_info,
            ca_file=ca_file,
            setup_pkinit=not options.no_pkinit,
        )
    ds.master_fqdn = config.master_host_name
    if ca_is_configured is not None:
        ds.ca_is_configured = ca_is_configured
    ds.promote = promote
    ds.api = remote_api

    # from __setup_replica

    # Always connect to ds over ldapi
    ldap_uri = ipaldap.get_ldap_uri(protocol='ldapi', realm=ds.realm)
    conn = ipaldap.LDAPClient(ldap_uri)
    conn.external_bind()

    return ds


def ansible_module_get_parsed_ip_addresses(ansible_module,
                                           param='ip_addresses'):
    ip_addrs = []
    for ip in ansible_module.params.get(param):
        try:
            ip_parsed = ipautil.CheckedIPAddress(ip)
        except Exception as e:
            ansible_module.fail_json(msg="Invalid IP Address %s: %s" % (ip, e))
        ip_addrs.append(ip_parsed)
    return ip_addrs


def gen_remote_api(master_host_name, etc_ipa):
    ldapuri = 'ldaps://%s' % ipautil.format_netloc(master_host_name)
    xmlrpc_uri = 'https://{}/ipa/xml'.format(
        ipautil.format_netloc(master_host_name))
    remote_api = create_api(mode=None)
    remote_api.bootstrap(in_server=True,
                         context='installer',
                         confdir=etc_ipa,
                         ldap_uri=ldapuri,
                         xmlrpc_uri=xmlrpc_uri)
    remote_api.finalize()
    return remote_api
